#include "cthugha.h"
#include "display.h"
#include "options.h"
#include "keys.h"
#include "action.h"
#include "information.h"
#include "sound.h"
#include "translate.h"

#include <string.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#undef COLORS					/* comes again in ncurses */
#include <ncurses.h>

#ifdef CTH_X11
#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
Status mit_shm;
XShmSegmentInfo shminfo;
XImage * image;
extern Window cth_window;
extern GC cth_gc;
#endif

/*
 * This file is used for the "normal" cthugha-L and for the cthugha-X11.
 *
 * If CTH_X11 is defined, then the version for X11 should be generated.
 * If CTH_SS is also defined, then the screen saver version.
 */

/* possible display-function */
int screen_normal();	int screen_down();	int screen_2hor();
int screen_r2hor();	int screen_4hor();	int screen_2verd();
int screen_r2verd();	int screen_4kal();	int screen_mirrorH();
int screen_mirrorHV();	int screen_mirrorV();

opt_data screens[] = {
    { screen_normal,	1,	"Up",		"Normal Display" },
    { screen_down,	1,	"Down",		"Upside Down"},
    { screen_2hor,	1,	"2hor",		"Hor. Split out"},
    { screen_r2hor,	1,	"r2hor",	"Hor. Split in"},
    { screen_4hor,	1,	"4hor",		"Kaleidoscope"},
    { screen_2verd,	1,	"2verd",	"90deg rot. mirror"},
    { screen_r2verd,	1,	"r2verd",	"90deg rot. mirror II"},
    { screen_4kal,	1,	"4kal",		"90deg Kaleidoscope"},
    { screen_mirrorH,	1,	"mirrorH",	"Mirror horizontal"},
    { screen_mirrorHV,	1,	"mirrorHV",	"Mirror horiz&vert"},
    { screen_mirrorV,	1,	"mirrorV",	"Mirror vertical"}
};
int nr_screens = sizeof(screens) / sizeof(opt_data);
int screen_sizes[][2] = {
    {1,1},	/* normal */
    {1,1},	/* down */
    {1,1},	/* 2hor */
    {1,1},	/* r2hor */
    {1,1},	/* 4hor */
    {1,1},	/* 2verd */
    {1,1},	/* r2verd */
    {1,1},	/* 4 kal */
    {2,1},	/* mirrorH */
    {2,2},	/* mirrorHV */
    {1,2}	/* mirrorV */
};

int (*update_screen)(void);			/* pointer to display-funct. */

int screen_width = 0;
int screen_height = 0;

#define	SCREEN_SIZE	( screen_width * screen_height )

char * buff1;					/* Screen-Memory 1 */
char * buff2;					/* Screen-Memory 2 */

unsigned char * active_buffer;			/* Next on screen */
unsigned char * passive_buffer;			/* right now on screen */

char screen_first[256] = "";			/* Start with this scrn-fkt */

int display_frames = 0;				/* counter for speed-tests */
time_t display_start;

int display_clear = 1;				/* clear screen */

int display_tile_x = 1;				/* tile smal buffers on scrn */
int display_tile_y = 1;
static int tiles_x = 1, tiles_y = 1;		/* how many tiles */
int display_mode = 0;				/* graphic mode */
int screen_modes[] = {				/* allowed modes */
    G320x200x256,		/* 0 */
    G320x240x256,		/* 1 */
    G320x400x256,		/* 2 */
    G360x480x256,		/* 3 */
    G640x480x256,		/* 4 */
    G800x600x256,		/* 5    here it starts getting tiny */
    G1024x768x256,		/* 6  */
    G1280x1024x256		/* 7    only for completeness */
};
int nr_screen_modes = sizeof(screen_modes) / sizeof(int);

unsigned char * display_mem = NULL;		/* memory for screen-fkt
						   to draw to */
unsigned char * display_bitmap = NULL;		/* memory for X11-bitmap */
int display_direct = 1;				/* draw directly to screen */
int display_copybox = 0;			/* use copybox. Not available
						   at some screen-modes */

int display_text_line=-1;			/* display at line */
char display_text[MAX_TEXT_DISPLAY+1];		/* text to display */
char display_font[ 256 * 8 * 8 ];
int display_text_on_screen = 0;			/* text is currently on screen;
						   also used to count how long
						   the text is on the screen 
						   (counts down, on 0 the text
						   is removed	*/
int display_syncwait = 
#ifndef CTH_X11
	0;			/* Wait for sync on display */
#else
	1;			/* call XSync */
#endif
GraphicsContext display_screen_context;
GraphicsContext	display_virtual_context;

int display_mit_shm = 1;			/* use MIT-SHM if possible */
int display_on_root = 0;			/* display on root window */
int window_width;
int window_height;
#ifndef CTH_SS
/* 
 *  Initialization 
 */
int init_ncurses() {
    if( !initscr() )
	return 1;

    timeout(0);					/* no timeout on getch(); */
    keypad(stdscr, TRUE);			/* allow function keys */
    meta(stdscr, TRUE);				/* 8-bit clean */
    noecho();					/* don't print out keys */
    cbreak();					/* don't wait for cr */
    start_color();

    clear();
    refresh();
    
    return 0;
}
#endif

int init_display() {
  
#ifdef CTH_X11
    static int X[] = {320,320,320,260,640,800,1024,1280};
    static int Y[] = {200,240,400,480,480,600,768,1024};
  
    if ( display_mode == -1) {
	/* a special display resolution is given */
	tiles_x = (display_tile_x ? (screen_width/BUFF_WIDTH) : 1);
	tiles_y = (display_tile_y ? (screen_height/(BUFF_HEIGHT-4)):1);
    } else {
	/* use one of the default resolutions */
	if ( (display_mode >= nr_screen_modes) || (display_mode < 0))
	    display_mode = 0;
	
	tiles_x = (display_tile_x ?(X[display_mode]/BUFF_WIDTH) : 1);
	tiles_y = (display_tile_y ?(Y[display_mode]/(BUFF_HEIGHT-4)):1);
    }
    /* make buffer big enough */
    tiles_x = max(tiles_x,1);
    tiles_y = max(tiles_y,1);

    /* calculate real size of display */
    screen_width = max(screen_width, BUFF_WIDTH * tiles_x);
    screen_height = max(screen_height, (BUFF_HEIGHT-4) * tiles_y);

    /* Check if MIT-SHM is available */
    if ( display_mit_shm) {
	int d1,d2;    Bool d3;		/* dummy variables */
	mit_shm = XShmQueryVersion(cth_display,&d1,&d2,&d3);
    }

    alloc_image();
#else
    vga_init();					/* initialize svgalib */
#endif
    
    /* allocate memory for buffers */
    if( (buff1 = malloc(BUFF_SIZE+3*BUFF_WIDTH)) == NULL) {
	printfe("Can not allocate memory for buffer 1.\n");
	return 1;
    }
    if( (buff2 = malloc(BUFF_SIZE+3*BUFF_WIDTH)) == NULL) {
	printfe("Can not allocate memory for buffer 2.\n");
	return 1;
    }

    /* on top of buffer 2 lines not displayed for general flame functions */
    active_buffer = buff1 + 2*BUFF_WIDTH;
    passive_buffer = buff2 + 2*BUFF_WIDTH;
	
    /* clear active buffer */
    memset( active_buffer, BUFF_SIZE, 0);

    /* initiaize the palettes (internal and external) */
    if( init_palettes() )
	return 1;
		
    /* load the pcx-files */
    if ( init_pcx() )
	return 1;

    /* check if any palettes in use */
    if ( nr_palettes == 0) {
	printfe("No palettes specified.\n");
	return 1;
    }

    return 0;
}

#ifdef CTH_X11
/*
 * Allocate the XImage
 */
int alloc_image() {

    if ( mit_shm) {
	/* create XImage */
	image = XShmCreateImage(cth_display,
				DefaultVisual(cth_display, 
					      DefaultScreen(cth_display)),
				8,		/* 8 bpp */
				ZPixmap,	/* format */
				NULL,		/* data */
				&shminfo, 
				screen_width, screen_height);
	if ( image == NULL) {
	    printfe("Can not create XImage.\n");
	    return 1;
	}

	/* maybe the size changed */
	screen_width = image->bytes_per_line;
	screen_height = image->height;

	/* create Shared Memory */
	if ( (shminfo.shmid = shmget(IPC_PRIVATE, SCREEN_SIZE,
				     IPC_CREAT|0777)) == -1 ) {
	    printfee("Can not create shared memory segment");
	    return 1;
	}
	/* attach Shard Memory */
	if( (shminfo.shmaddr = image->data = shmat(shminfo.shmid, 0,0)) 
	   == (void*)-1) {
	    printfee("Can not attach shared memory segment");
	    return 1;
	}
	    
	/* allow X11 to write */
	shminfo.readOnly = False;

	/* Attach X11 with Shared Memory */
	if( XShmAttach(cth_display, &shminfo) == 0) {
	    printfe("Can not X-attach shared memory segment.\n");
	    return 1;
	}

	/* keep pointer to data */
	display_bitmap = image->data;

    } else {
	/* allocate memory for bitmap without Shard Memory*/
	if( (display_bitmap = (unsigned char *)malloc( SCREEN_SIZE)) == NULL) {
	    printfe("Can not allocate memory for bitmap.\n");
	    return 1;
	}
	/* create Image */
	image = XCreateImage(cth_display, 
			     DefaultVisual(cth_display, 
					   DefaultScreen(cth_display)),
			     8, ZPixmap, 0, 
			     display_bitmap, screen_width, screen_height,
			     XBitmapPad(cth_display),  screen_width);
	if ( image == NULL) {
	    printfe("Can not create XImage.\n");
	    return 1;
	}

	/* maybe the size changed */
	screen_width = image->bytes_per_line;
	screen_height = image->height;
    }
    return 0;
}

/*
 * Resize window
 */
int resize_display(int new_width, int new_height) {

    /* calulcate new real size of display */
    tiles_x = (display_tile_x ? (new_width/BUFF_WIDTH) : 1);
    tiles_y = (display_tile_y ? (new_height/(BUFF_HEIGHT-4)):1);

    /* make buffer big enough */
    tiles_x = max(tiles_x,1);
    tiles_y = max(tiles_y,1);

    /* calculate real size of display */
    screen_width = max(screen_width, BUFF_WIDTH * tiles_x);
    screen_height = max(screen_height, (BUFF_HEIGHT-4) * tiles_y);

    /* release the old image */
    if( free_image() )
	return 1;

    /* allocate a new image */
    if( alloc_image() )
	return 1;

    return 0;
}
#endif

#ifndef CTH_SS
/*
 * Exit-stuff
 */
int exit_ncurses() {
    echo();
    endwin();
    return 0;
}
#endif

int exit_display() {
    display_text_mode();
#ifdef CTH_X11
    free_image();
#endif
    return 0;
}

#ifdef CTH_X11
/*
 * free the XImage
 */
int free_image() {
    if ( mit_shm) {
	/* remove Shared Memory if needed */
	XShmDetach(cth_display, &shminfo);
	XDestroyImage(image);
	shmdt(shminfo.shmaddr);
	shmctl(shminfo.shmid,IPC_RMID,0);
    } else {
	XDestroyImage(image);
    }
    return 0;
}
#endif

/*
 * Switch to ext mode
 */
int display_text_mode() {
#ifndef CTH_X11
    vga_setmode(TEXT);
#endif 
    
    return 0;
}

/*
 * Enable graph-mode
 */
#ifndef CTH_SS
int display_graph_mode() {

#ifndef CTH_X11
    vga_modeinfo * mi;

    if ( (display_mode >= nr_screen_modes) || (display_mode < 0))
	display_mode = 0;

    /* create screen-context */
    vga_setmode(screen_modes[ display_mode ]);		
    gl_setcontextvga(screen_modes[ display_mode ]);
    gl_getcontext(&display_screen_context);
/*    gl_enablepageflipping( &display_screen_context);  */
    /* direct display only in 320x200 */
    if ( display_mode != 0)
	display_direct = 0;

    mi = vga_getmodeinfo(screen_modes[ display_mode ]);

    /* can't use copybox if screen-mode is planar 256 colors */
    display_copybox = (mi->flags & 4) ? 0 : 1;
 
    /* prepare font */
    gl_expandfont(8,8,255, gl_font8x8, display_font);
    gl_setfont( 8,8, display_font);
    gl_setwritemode( WRITEMODE_MASKED);
    gl_colorfont(8,8, 255, display_font);				

    /* create context for drawing in memory */
    gl_setcontextvgavirtual(screen_modes[ display_mode ]);
    gl_getcontext(&display_virtual_context);
/*    gl_enablepageflipping( &display_virtual_context);  */

    screen_width = WIDTH;			/* from SVGALIB */
    screen_height = HEIGHT;			/* from SVGALIB */

    if ( screen_width < BUFF_WIDTH) {
	display_text_mode();
	printfe("Screen must not be smaller than buffer.");
	return 1;
    }

    tiles_x = (display_tile_x ? (screen_width/BUFF_WIDTH) : 1);
    tiles_y = (display_tile_y ? (screen_height/(BUFF_HEIGHT-4)) : 1);

    tiles_x = max(tiles_x, 1);
    tiles_y = max(tiles_y, 1);

#else
    clear();
    refresh();
#endif

    update_palette();				/* set current palette */

    display_start = time(0);
	
    return 0;
}
#endif

/*
 * Change/Select how buffer is displayed on screen
 */
int change_update_screen(int to) {
    update_screen = opt_change(to, screens, nr_screens, update_screen)->data;
#ifndef CTH_SS
    update_status();
#endif
    display_clear = 1;
    return 0;
}
int select_screen() {
    int i;
    if( (i=display_selection(screens, nr_screens, 
			     update_screen, "Select Screens\n")) >= 0) 
	change_update_screen(i);
    return i;
}

/* Display a text on the screen */
/* modified to allow status to change without reseting the counter...  clp */
int display_print(char * text, int line, int reset) {

#ifndef CTH_SS

    clear_status();				/* status is no longer disp. */
#ifndef CTH_X11
    if ( text == NULL) {			/* deactivate display */
	if ( display_text_on_screen) {		/* something on display */
	    display_text_on_screen = 0;
	    update_palette();
	    display_clear = 1;
	}
    } else {					/* disp. something new */
	strncpy( display_text, text, MAX_TEXT_DISPLAY);	/* copy text */
	strcat( display_text, "\n");		/* makes display simplier */
	if ( line < 0)				/* determine line on display */
	    display_text_line = rand() % (screen_height - 8);
	else
	    display_text_line = line;

	if ( reset )
	    display_text_on_screen = 255;
	update_palette();
    }
#else
    if ( text == NULL) 
	display_text[0] = '\0';
    else
	strncpy(display_text, text, MAX_TEXT_DISPLAY);
    erase();
    mvprintw(0,0, display_text);
    refresh();
#endif

#endif /* CTH_SS */

    return 0;
}

#ifndef CTH_SS
/*
 * Selection-stuff
 */
int sel_line = 0;				/* nr. of line cur. selected */
opt_data * sel_what = NULL;
int sel_nr = 0;
char * sel_title = NULL;

char * name_dest_str(char * name, char * desc) {
    static char str[MAX_NAME_LEN + MAX_DESC_LEN + 1];

    sprintf(str, "%-8s %25s", name, desc);
    str[34] = '\0';
    return str;
}

int display_selection(opt_data * what, int nr, void * old, char * title) {
    int s;

    sel_what = what;
    sel_nr = nr;
    sel_title = title;
    sel_line = opt_number(old,what,nr);

#ifndef CTH_X11
    while( (s=display_do_selection()) == -2);
#else
    s=display_do_selection();
#endif

    return s;
}

#define SEL_SIZE	5
int display_do_selection() {
    char s[256];				/* temporary string */
    char text[20000];				/* text disp. on display */

    int i, key;

    if ( sel_nr <= 0)
	return -2;

    /* create text for selection */
    strcpy(text, sel_title);
    for(i= sel_line - SEL_SIZE; i < sel_line + SEL_SIZE; i++) {
	if ( i < 0)
	    sprintf(s,"\n");
	else if ( i == sel_line)
	    sprintf(s, ">%34s%4s<\n", 
		    name_dest_str(sel_what[i].name, sel_what[i].desc),
		    sel_what[i].use ? "ON" : "OFF");
	else if ( i < sel_nr)
	    sprintf(s, " %34s%4s \n", 
		    name_dest_str(sel_what[i].name, sel_what[i].desc),
		    sel_what[i].use ? "ON" : "OFF");
	else
	    s[0] = '\0';
					
	strcat( text, s);
    }
    
    display_print(text, 0,1);
    
#ifndef CTH_X11	
    /* business as usual */
    display_sound();
#endif
    
    while( (key = getkey() ) != Z_NOKEY  ) {
	switch( key ) {
	case Z_UP:
	    if (sel_line > 0)
		sel_line --;
	    break;
	case Z_DOWN:
	    if (sel_line < (sel_nr-1))
		sel_line ++;
	    break;
	case Z_PGUP:
	    sel_line -= 2*SEL_SIZE-1;
	    if ( sel_line < 0)
		sel_line = 0;
	    break;
	case Z_PGDN:
	    sel_line += 2*SEL_SIZE+1;
	    if ( sel_line >= (sel_nr-1))
		sel_line = sel_nr -1;
	    break;
	case Z_HOME:
	    sel_line = 0;
	    break;
	case Z_END:
	    sel_line = sel_nr -1;
	    break;
	case Z_SPACE:
	    sel_what[sel_line].use = 1 - sel_what[sel_line].use;
	    break;
	case Z_ENTER:
	    sel_what[sel_line].use = 1;
	    display_print(NULL, 0,1);
	    sel_nr = 0;
	    return sel_line;
/*	default: */
	case Z_ESC:
	    display_print(NULL, 0,1);
	    sel_nr = 0;
	    return -1;
	}
    }
    return -2;
}
#else
int display_selection(opt_data * what, int nr, void * old, char * title) {
    return opt_number(old,what,nr);;
}
#endif /* CTH_SS */    	

/*****************************************************************************/

/*
 * exchange active and passive buffer
 */
void swap_buffers() {
    char * tmp;
    tmp = active_buffer; active_buffer = passive_buffer; passive_buffer = tmp;
}

/*
 * bring the current active text to the screen
 */
void show_text() {

#ifndef CTH_X11

    if ( ! display_text_on_screen ) 
	return;
		
    display_text_on_screen --;			/* decrease counter */

    /* keep on screen if status */
    if ( status_active() && (display_text_on_screen <= 0)) {
	display_text_on_screen = 255;
    }
    /* check, if text should be removed */
    if ( display_text_on_screen <= 0) {	
	display_text_on_screen = 0;
	update_palette();				/* reset palette */
	display_clear = 1;
    } else {
	/* bring text to screen (line by line) */
	char * text = display_text;
	char * ntext = strchr(text,'\n');
	int line = display_text_line;
	while ( ntext != NULL) {
	    gl_writen( 0,line, min(ntext-text, screen_width/8), text);
	    line += 8;
	    if ( line >= screen_height)
		line = 0;
	    text = ntext+1;
	    ntext = strchr(text, '\n');
	}
    }
#endif
}

/*****************************************************************************
 some useful definitions
*/

#define SB_HEIGHT	( min(BUFF_HEIGHT-4, screen_height) )

#define SCREEN_OFFSET_X	(max(((screen_width -BUFF_WIDTH*tile_x*size_x) / 2),0))
#define SCREEN_OFFSET_Y	(max(((screen_height-SB_HEIGHT *tile_y*size_y) / 2),0))
#define SCREEN_OFFSET	(screen_width * SCREEN_OFFSET_Y + SCREEN_OFFSET_X)

#define SB_HEIGHT2	(SB_HEIGHT/2)
#define SB_SIZE2	(SB_HEIGHT2 * BUFF_WIDTH)

#define BUFF_WIDTH2	(BUFF_WIDTH/2)
#define screen_width2	(screen_width/2)

/*
 * clear border around display if necessary 
 */
int clear_border(int tile_x,int tile_y, int size_x,int size_y) {

#ifdef CTH_X11
    int i;
    if ( display_clear ) {
	/* upper border */
	memset(display_bitmap, 0, screen_width*SCREEN_OFFSET_Y);
	/* left & right border */
	for(i=SCREEN_OFFSET_Y; i < screen_height-SCREEN_OFFSET_Y; i++) {
	    memset(display_bitmap + i*screen_width, 0, SCREEN_OFFSET_X);
	    memset(display_bitmap + i*screen_width 
		   + screen_width - SCREEN_OFFSET_X, 0, SCREEN_OFFSET_X);
	}
	/* lower border */
	memset(display_bitmap+(screen_height-SCREEN_OFFSET_Y)*screen_width, 
	       0,screen_width * SCREEN_OFFSET_Y );

	display_clear = 0;

	return 1;
    }
#else
    if ( display_text_on_screen || display_clear ) {
	gl_fillbox(0,0, 
		   screen_width, SCREEN_OFFSET_Y,
		   0);
	gl_fillbox(0,SCREEN_OFFSET_Y,	
		   SCREEN_OFFSET_X, SB_HEIGHT*tile_y*size_y,	
		   0);
	gl_fillbox(screen_width-SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
		   SCREEN_OFFSET_X, SB_HEIGHT*tile_y*size_y,
		   0);
	gl_fillbox(0,screen_height-SCREEN_OFFSET_Y,
		   screen_width, SCREEN_OFFSET_Y,			
		   0);

	display_clear = 0;

	return 1;
    }
#endif
    return 0;
}

/*
 * tile buffer to screen
 */
int tile_buffer(int tile_x,int tile_y, int size_x,int size_y) {
#ifdef CTH_X11
    int i,j;
    
    for(i=1; i < tile_x; i++)
	for(j=0; j < (BUFF_HEIGHT-4)*size_y; j++)
	    memcpy(display_mem + (i*BUFF_WIDTH*size_x + j*screen_width),
		   display_mem + (j*screen_width),
		   BUFF_WIDTH*size_x);
    
    for(i=1; i < tile_y; i++)
	for(j=0; j < (BUFF_HEIGHT-4)*size_y; j++)
	    memcpy(display_mem + (i*screen_width*(BUFF_HEIGHT-4)*size_y
				  +j*screen_width),
		   display_mem + (j*screen_width),
		   screen_width);
    
#else
    int i;

    for(i=1; i < tile_x; i++)
    	gl_copybox( SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
    		    BUFF_WIDTH*size_x, SB_HEIGHT*size_y,
    		    SCREEN_OFFSET_X + i*BUFF_WIDTH*size_x, SCREEN_OFFSET_Y);
    		    
    for(i=1; i < tile_y; i++)
    	gl_copybox( SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
    		    BUFF_WIDTH*tile_x*size_x, SB_HEIGHT*size_y,
    		    SCREEN_OFFSET_X, SCREEN_OFFSET_Y + i*SB_HEIGHT*size_y);
    		    
#endif 
    return 0;
}

/*
 * calculate information for tile and offset 
 */
int get_tile_size(int * tile_x, int * tile_y, int * size_x, int * size_y) {

    *size_x = screen_sizes[ opt_number(update_screen, screens, nr_screens)][0];
    *tile_x = max(tiles_x / *size_x, 1);
	
    *size_y = screen_sizes[ opt_number(update_screen, screens, nr_screens)][1];
    *tile_y = max(tiles_y / *size_y, 1);

    return 0;
}

#ifdef CTH_X11
#define display_base display_bitmap
#else
#define display_base VBUF
#endif
/*
 * Bring the active Buffer to screen.
 */
int display() {
    int tile_x, tile_y, size_x, size_y;
    int clear;
    
#ifndef CTH_X11	
    if( display_direct )
	gl_setcontext( &display_screen_context);
    else
	gl_setcontext( &display_virtual_context);
#endif

    do {
	get_tile_size(&tile_x, &tile_y, &size_x, &size_y);
	display_mem = display_base + SCREEN_OFFSET;
    } while( update_screen() );			/* draw the buffer */

    clear = clear_border(tile_x,tile_y, size_x,size_y);	 
    tile_buffer(tile_x,tile_y, size_x,size_y);	
    show_text();				/* bring active text to scrn */

#ifndef CTH_X11					/* normal mode */
    if( display_syncwait )
	vga_waitretrace();

    /* copy virtual screen to physical screen */
    if( !display_direct) {
	if( clear || !display_copybox)		/* draw full screen */
	    gl_copyscreen( &display_screen_context);
	else {					/* or only a part */
	    gl_setcontext( &display_screen_context);
	    gl_copyboxfromcontext( &display_virtual_context,
				  SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
				  BUFF_WIDTH * tile_x * size_x, 
				  SB_HEIGHT * tile_y * size_y,
				  SCREEN_OFFSET_X, SCREEN_OFFSET_Y);
	}
    }
#else						/* X11 mode */
    {
	int ss_offset_x = max((window_width - screen_width) >> 1, 0);
	int ss_offset_y = max((window_height - screen_height) >> 1, 0);

	if(clear)
	    if ( mit_shm) {
		XShmPutImage(cth_display, cth_window, cth_gc, image,0,0,
			     ss_offset_x,  ss_offset_y, 
			     screen_width, screen_height, 
			     0);
	    } else {
		XPutImage(cth_display, cth_window, cth_gc, image, 0,0,
			  ss_offset_x, ss_offset_y,
			  screen_width, screen_height);
	    }
	else
	    if ( mit_shm) {
		XShmPutImage(cth_display, cth_window, cth_gc, image,
			     SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			     ss_offset_x + SCREEN_OFFSET_X,
			     ss_offset_y + SCREEN_OFFSET_Y,
			     BUFF_WIDTH * tile_x * size_x,
			     SB_HEIGHT * tile_y * size_y, 
			     0);
	    } else {
		XPutImage(cth_display, cth_window, cth_gc, image, 
			     SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			     ss_offset_x + SCREEN_OFFSET_X,
			     ss_offset_y + SCREEN_OFFSET_Y,
			     BUFF_WIDTH * tile_x * size_x,
			     SB_HEIGHT * tile_y * size_y);
	    }
    

	if ( display_syncwait )
	    XSync(cth_display, 0);
    }
#endif

    swap_buffers();
    display_frames ++;

    return 0;
}


/*****************************************************************************
 * Screen-functions
 ****************************************************************************/

/*
 * screen functions with size 1x1
 */

int screen_normal() {
    int i;
    char * scrn, * tmp;
		
    scrn = display_mem;
    tmp = active_buffer;
    for(i=SB_HEIGHT; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);
	scrn += screen_width;
	tmp += BUFF_WIDTH;
    }
    return 0;
}

int screen_down() {
    char * tmp, * scrn;
    int i;
    
    tmp = active_buffer;
    scrn = display_mem + screen_width * (SB_HEIGHT-1);
    for(i=SB_HEIGHT; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);
	tmp += BUFF_WIDTH;
	scrn -= screen_width;
    }
    return 0;
}

int screen_2hor() {
    char * tmp, * scrn;
    int i;
	
    /* lower half of buffer maps to upper half of screen */

    tmp = active_buffer + SB_SIZE2;
    scrn = display_mem;
				
    for(i=SB_HEIGHT2; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);
	scrn += screen_width;
	tmp += BUFF_WIDTH;
    }
		
    /* lower half of buffer get turned around to lower half of screen*/

    tmp = active_buffer + SB_SIZE2;
    scrn = display_mem + screen_width*(SB_HEIGHT-1);

    for(i=SB_HEIGHT2; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);			
	tmp += BUFF_WIDTH;
	scrn -= screen_width;
    }
    return 0;
}
		
int screen_r2hor() {
    char * tmp, * scrn;
    int i;
	
    /* lower half of buffer maps to lower half of screen */

    tmp = active_buffer + SB_SIZE2;
    scrn = display_mem + SB_HEIGHT2*screen_width;
    
    for(i=SB_HEIGHT2; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);
	scrn += screen_width;
	tmp += BUFF_WIDTH;
    }				
		
    /* lower half of buffer get turned around to upper half of screen*/
	
    tmp = active_buffer + SB_SIZE2;
    scrn = display_mem + (SB_HEIGHT2-1)*screen_width;
    for(i=SB_HEIGHT2; i != 0; i--) {
	memcpy( scrn, tmp, BUFF_WIDTH);			
	tmp += BUFF_WIDTH;
	scrn -= screen_width;
    }
    return 0;
}
		
int screen_4hor() {
    char * tmp, * scrn;
    int x,y;
	
    /* upper half of screen */	
	
    tmp = active_buffer + SB_SIZE2;
    scrn = display_mem;

    for(y=SB_HEIGHT2; y != 0; y--) {
	
	/* left half */
	memcpy( scrn, tmp, BUFF_WIDTH2);				
	scrn += BUFF_WIDTH2;
	tmp += BUFF_WIDTH2;
		
	/* right half */
	for(x=BUFF_WIDTH2; x != 0; x--)	{
	    *scrn = * tmp;
	    scrn ++;
	    tmp --;
	}
	tmp += BUFF_WIDTH;
	scrn += screen_width - BUFF_WIDTH;
    }

    /* lower half of screen */
	
    tmp = active_buffer + BUFF_WIDTH*SB_HEIGHT - BUFF_WIDTH;
    scrn = display_mem + screen_width*SB_HEIGHT2;
	
    for(y= SB_HEIGHT2; y != 0; y--) {
		
	/* left half */
	memcpy( scrn, tmp, BUFF_WIDTH2);
	scrn += BUFF_WIDTH2;
	tmp += BUFF_WIDTH2;
		
	/* right half */
	for(x=BUFF_WIDTH2; x != 0; x--)	{
	    *scrn = * tmp;
	    scrn ++;
	    tmp --;
	}
	tmp -= BUFF_WIDTH;
	scrn += screen_width - BUFF_WIDTH;
    }
    return 0;
}

int screen_2verd() {
    char * tmp, * scrn;
    int x,y;

    if( BUFF_WIDTH2 <= BUFF_HEIGHT) {
	
	tmp = active_buffer;
	scrn = display_mem;
	
	for(y=SB_HEIGHT; y != 0; y--) {
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn ++ = *tmp;
		tmp += BUFF_WIDTH;
	    }
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn ++ = *tmp;
		tmp -= BUFF_WIDTH;
	    }
	    tmp ++;
	    scrn += screen_width - BUFF_WIDTH;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}

int screen_r2verd() {
    char * tmp, * scrn;
    int x,y;
	
    if( BUFF_WIDTH2 <= BUFF_HEIGHT) {
	tmp = active_buffer;
	scrn = display_mem + screen_width * (SB_HEIGHT-1) + BUFF_WIDTH -1;
	
	for(y=SB_HEIGHT; y != 0; y--) {
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn -- = *tmp;
		tmp += BUFF_WIDTH;
	    }
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn -- = *tmp;
		tmp -= BUFF_WIDTH;
	    }
	    tmp ++;
	    scrn -= screen_width - BUFF_WIDTH;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}

int screen_4kal() {
    char * tmp, * scrn;
    int x,y;
	
    if( BUFF_WIDTH2 <= BUFF_HEIGHT) {
	tmp = active_buffer;
	scrn = display_mem;
	
	/* upper half */
	for(y=SB_HEIGHT2; y != 0; y--) {
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn ++ = *tmp;
		tmp += BUFF_WIDTH;
	    }
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn ++ = *tmp;
		tmp -= BUFF_WIDTH;
	    }
	    tmp ++;
	    scrn += screen_width - BUFF_WIDTH;
	}
	
	tmp = active_buffer;
	scrn = display_mem + screen_width * SB_HEIGHT - screen_width + 
	    BUFF_WIDTH - 1;
	
	/* lower half */
	for(y=SB_HEIGHT2; y != 0; y--) {
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn -- = *tmp;
		tmp += BUFF_WIDTH;
	    }
	    for(x=BUFF_WIDTH2; x != 0; x--) {
		*scrn -- = *tmp;
		tmp -= BUFF_WIDTH;
	    }
	    tmp ++;
	    scrn -= screen_width - BUFF_WIDTH;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}

/*
 * big screen functions
 */

int screen_mirrorH() {
    int i,j;
    char * src;
    char * dst1, * dst2;
		
    if( 2*BUFF_WIDTH <= screen_width) {

	src = active_buffer;
	dst1 = display_mem;
	dst2 = display_mem + 2*BUFF_WIDTH-1;

	for(i=SB_HEIGHT; i != 0; i--) {
	    /* left part maps directly */
	    memcpy(dst1, src, BUFF_WIDTH);
	    
	    /* right part gets turned around */
	    for(j=BUFF_WIDTH; j != 0; j--) {
		*dst2 = *src;
		src ++;
		dst2 --;
	    }
	    /* move pointer to next line */
	    dst1 += screen_width;
	    dst2 += screen_width + BUFF_WIDTH;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}

int screen_mirrorHV() {
    int i,j;
    char * src;
    char * dst1, * dst2, *dst3, *dst4;
		
    if( (2*BUFF_WIDTH <= screen_width) && (2*SB_HEIGHT <= screen_height) ) {

	src = active_buffer;
	dst1 = display_mem;
	dst2 = display_mem + 2*BUFF_WIDTH-1;
	dst3 = display_mem + 2*SB_HEIGHT*screen_width - screen_width;
	dst4 = display_mem + 2*SB_HEIGHT*screen_width + 2*BUFF_WIDTH-1
	    - screen_width;

	for(i=SB_HEIGHT; i != 0; i--) {
	    /* upper left maps direct */
	    memcpy(dst1, src, BUFF_WIDTH);
	    /* lower left maps direct */
	    memcpy(dst3, src, BUFF_WIDTH);
	    /* upper & lower right gets turned around */
	    for(j=BUFF_WIDTH; j != 0; j--) {
		*dst2 = *src;
		*dst4 = *src;
		src ++;
		dst2 --;
		dst4 --;
	    }
	    /* move pointer to next line */
	    dst1 +=  screen_width;
	    dst2 +=  screen_width + BUFF_WIDTH;
	    dst3 += -screen_width;
	    dst4 += -screen_width + BUFF_WIDTH;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}

int screen_mirrorV() {
    int i;
    char * src;
    char * dst1, * dst2;
		
    if( 2*SB_HEIGHT <= screen_height ) {

	src = active_buffer;
	dst1 = display_mem;
	dst2 = display_mem + 2*SB_HEIGHT*screen_width - screen_width;

	for(i=SB_HEIGHT; i != 0; i--) {
	    /* do 1 line */
	    memcpy(dst1, src, BUFF_WIDTH);
	    memcpy(dst2, src, BUFF_WIDTH);
	    /* move pointer to next line */
	    src += BUFF_WIDTH;
	    dst1 += screen_width;
	    dst2 -= screen_width;
	}
    } else {
	change_update_screen(CHANGE_NEXT);
	return 1;
    }
    return 0;
}





